@@ -9,6 +9,7 @@
 #ifndef Z_SOLO
 #  include "gzguts.h"
 #endif
+
 /* Added by Yixuan Qiu */
 /* Use error handler in R */
 #include <R_ext/Error.h>
@@ -17,7 +18,7 @@
 struct internal_state      {int dummy;}; /* for buggy compilers */
 #endif
 
-const char * const z_errmsg[10] = {
+z_const char * const z_errmsg[10] = {
 "need dictionary",     /* Z_NEED_DICT       2  */
 "stream end",          /* Z_STREAM_END      1  */
 "",                    /* Z_OK              0  */
